"""Tests for Fiona's OGR driver interface."""


import logging

import pytest

import fiona
from fiona.errors import FionaDeprecationWarning


def test_options(tmpdir, path_coutwildrnp_shp):
    """Test that setting CPL_DEBUG=ON works and that a warning is raised."""
    logfile = str(tmpdir.mkdir('tests').join('test_options.log'))
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    fh = logging.FileHandler(logfile)
    fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)

    # fiona.drivers() will be deprecated.
    with pytest.warns(FionaDeprecationWarning):
        with fiona.drivers(CPL_DEBUG=True):
            c = fiona.open(path_coutwildrnp_shp)
            c.close()
            with open(logfile, "r") as f:
                log = f.read()
            if fiona.gdal_version.major >= 2:
                assert "GDALOpen" in log
            else:
                assert "OGROpen" in log



"""$ fio bounds"""


import json
import logging

import click
from cligj import precision_opt, use_rs_opt

import fiona
from fiona.fio.helpers import obj_gen
from fiona.fio import with_context_env


@click.command(short_help="Print the extent of GeoJSON objects")
@precision_opt
@click.option('--explode/--no-explode', default=False,
              help="Explode collections into features (default: no).")
@click.option('--with-id/--without-id', default=False,
              help="Print GeoJSON ids and bounding boxes together "
                   "(default: without).")
@click.option('--with-obj/--without-obj', default=False,
              help="Print GeoJSON objects and bounding boxes together "
                   "(default: without).")
@use_rs_opt
@click.pass_context
@with_context_env
def bounds(ctx, precision, explode, with_id, with_obj, use_rs):
    """Print the bounding boxes of GeoJSON objects read from stdin.

    Optionally explode collections and print the bounds of their
    features.

    To print identifiers for input objects along with their bounds
    as a {id: identifier, bbox: bounds} JSON object, use --with-id.

    To print the input objects themselves along with their bounds
    as GeoJSON object, use --with-obj. This has the effect of updating
    input objects with {id: identifier, bbox: bounds}.
    """
    logger = logging.getLogger(__name__)
    stdin = click.get_text_stream('stdin')
    try:
        source = obj_gen(stdin)
        for i, obj in enumerate(source):
            obj_id = obj.get('id', 'collection:' + str(i))
            xs = []
            ys = []
            features = obj.get('features') or [obj]
            for j, feat in enumerate(features):
                feat_id = feat.get('id', 'feature:' + str(i))
                w, s, e, n = fiona.bounds(feat)
                if precision > 0:
                    w, s, e, n = (round(v, precision)
                                  for v in (w, s, e, n))
                if explode:
                    if with_id:
                        rec = {
                            'parent': obj_id,
                            'id': feat_id,
                            'bbox': (w, s, e, n)}
                    elif with_obj:
                        feat.update(parent=obj_id, bbox=(w, s, e, n))
                        rec = feat
                    else:
                        rec = (w, s, e, n)
                    if use_rs:
                        click.echo(u'\u001e', nl=False)
                    click.echo(json.dumps(rec))
                else:
                    xs.extend([w, e])
                    ys.extend([s, n])
            if not explode:
                w, s, e, n = (min(xs), min(ys), max(xs), max(ys))
                if with_id:
                    rec = {'id': obj_id, 'bbox': (w, s, e, n)}
                elif with_obj:
                    obj.update(id=obj_id, bbox=(w, s, e, n))
                    rec = obj
                else:
                    rec = (w, s, e, n)
                if use_rs:
                    click.echo(u'\u001e', nl=False)
                click.echo(json.dumps(rec))

    except Exception:
        logger.exception("Exception caught during processing")
        raise click.Abort()



"""Tests of the transform submodule"""

import math
import pytest
from fiona import transform


@pytest.mark.parametrize(
    "geom",
    [
        {"type": "Point", "coordinates": [0.0, 0.0, 1000.0]},
        {
            "type": "LineString",
            "coordinates": [[0.0, 0.0, 1000.0], [0.1, 0.1, -1000.0]],
        },
        {
            "type": "MultiPoint",
            "coordinates": [[0.0, 0.0, 1000.0], [0.1, 0.1, -1000.0]],
        },
        {
            "type": "Polygon",
            "coordinates": [
                [
                    [0.0, 0.0, 1000.0],
                    [0.1, 0.1, -1000.0],
                    [0.1, -0.1, math.pi],
                    [0.0, 0.0, 1000.0],
                ]
            ],
        },
        {
            "type": "MultiPolygon",
            "coordinates": [
                [
                    [
                        [0.0, 0.0, 1000.0],
                        [0.1, 0.1, -1000.0],
                        [0.1, -0.1, math.pi],
                        [0.0, 0.0, 1000.0],
                    ]
                ]
            ],
        },
    ],
)
def test_transform_geom_with_z(geom):
    """Transforming a geom with Z succeeds"""
    g2 = transform.transform_geom("epsg:4326", "epsg:3857", geom, precision=3)


@pytest.mark.parametrize("crs", ["epsg:4326",
                                 "EPSG:4326",
                                 "WGS84",
                                 {'init': 'epsg:4326'},
                                 {'proj': 'longlat', 'datum': 'WGS84', 'no_defs': True},
                                 "OGC:CRS84"])
def test_axis_ordering(crs):
    """ Test if transform uses traditional_axis_mapping """

    expected = (-8427998.647958742, 4587905.27136252)
    t1 = transform.transform(crs, "epsg:3857", [-75.71], [38.06])
    assert (t1[0][0], t1[1][0]) == pytest.approx(expected)
    geom = {"type": "Point", "coordinates": [-75.71, 38.06]}
    g1 = transform.transform_geom(crs, "epsg:3857", geom, precision=3)
    assert g1["coordinates"] == pytest.approx(expected)

    rev_expected = (-75.71, 38.06)
    t2 = transform.transform("epsg:3857", crs, [-8427998.647958742], [4587905.27136252])
    assert (t2[0][0], t2[1][0]) == pytest.approx(rev_expected)
    geom = {"type": "Point", "coordinates": [-8427998.647958742, 4587905.27136252]}
    g2 = transform.transform_geom("epsg:3857", crs, geom, precision=3)
    assert g2["coordinates"] == pytest.approx(rev_expected)


def test_transform_issue971():
    """ See https://github.com/Toblerity/Fiona/issues/971 """
    source_crs = "epsg:25832"
    dest_src = "epsg:4326"
    geom = {'type': 'GeometryCollection', 'geometries': [{'type': 'LineString',
                                                          'coordinates': [(512381.8870945257, 5866313.311218272),
                                                                          (512371.23869999964, 5866322.282500001),
                                                                          (512364.6014999999, 5866328.260199999)]}]}
    geom_transformed = transform.transform_geom(source_crs, dest_src, geom, precision=3)
    assert geom_transformed['geometries'][0]['coordinates'][0] == pytest.approx((9.184, 52.946))



"""Tests of file opening"""

import io
import os

import pytest

import fiona
from fiona._crs import crs_to_wkt
from fiona.errors import DriverError

from .conftest import requires_gdal21


def test_open_shp(path_coutwildrnp_shp):
    """Open a shapefile"""
    assert fiona.open(path_coutwildrnp_shp)


def test_open_filename_with_exclamation(data_dir):
    path = os.path.relpath(os.path.join(data_dir, "!test.geojson"))
    assert os.path.exists(path), "Missing test data"
    assert fiona.open(path), "Failed to open !test.geojson"


@requires_gdal21
@pytest.mark.xfail(raises=DriverError)
def test_write_memfile_crs_wkt():
    example_schema = {
        "geometry": "Point",
        "properties": [("title", "str")],
    }

    example_features = [
        {
            "geometry": {"type": "Point", "coordinates": [0.0, 0.0]},
            "properties": {"title": "One"},
        },
        {
            "geometry": {"type": "Point", "coordinates": [1.0, 2.0]},
            "properties": {"title": "Two"},
        },
        {
            "geometry": {"type": "Point", "coordinates": [3.0, 4.0]},
            "properties": {"title": "Three"},
        },
    ]

    with io.BytesIO() as fd:
        with fiona.open(
            fd,
            "w",
            driver="GPKG",
            schema=example_schema,
            crs_wkt=crs_to_wkt("EPSG:32611"),
        ) as dst:
            dst.writerecords(example_features)

        fd.seek(0)
        with fiona.open(fd) as src:
            assert src.driver == "GPKG"
            assert src.crs == {"init": "epsg:32611"}



"""$ fio insp"""


import code
import logging
import sys

import click

import fiona
from fiona.fio import with_context_env


@click.command(short_help="Open a dataset and start an interpreter.")
@click.argument('src_path', required=True)
@click.option('--ipython', 'interpreter', flag_value='ipython',
              help="Use IPython as interpreter.")
@click.pass_context
@with_context_env
def insp(ctx, src_path, interpreter):
    """Open a collection within an interactive interpreter.
    """
    logger = logging.getLogger(__name__)
    banner = 'Fiona %s Interactive Inspector (Python %s)\n' \
             'Type "src.schema", "next(src)", or "help(src)" ' \
             'for more information.' \
             % (fiona.__version__, '.'.join(map(str, sys.version_info[:3])))

    try:
        with fiona.open(src_path) as src:
            scope = locals()
            if not interpreter:
                code.interact(banner, local=scope)
            elif interpreter == 'ipython':
                import IPython
                IPython.InteractiveShell.banner1 = banner
                IPython.start_ipython(argv=[], user_ns=scope)
            else:
                raise click.ClickException(
                    'Interpreter {} is unsupported or missing '
                    'dependencies'.format(interpreter))

    except Exception:
        logger.exception("Exception caught during processing")
        raise click.Abort()
